"""Container class for streams and functions."""
from __future__ import annotations

import typing

from ._all import secs_streams_functions
from .base import SecsStreamFunction

if typing.TYPE_CHECKING:
    import secsgem.common


class StreamsFunctions:
    """Container for functions classes."""

    def __init__(
        self,
        functions: list[type[SecsStreamFunction]] | None = None,
    ) -> None:
        """Initialize streams functions container."""
        if functions is None:
            functions = secs_streams_functions.copy()

        self._functions = functions

    def stream(self, stream: int) -> list[type[SecsStreamFunction]]:
        """Get all functions for a stream.

        Args:
            stream: number of the stream.

        Returns:
            list of function classes for this stream

        """
        return [function for function in self._functions if function.stream == stream]

    def function(self, stream: int, function: int) -> type[SecsStreamFunction] | None:
        """Get a specific function.

        Args:
            stream: number of the stream
            function: number of the function

        Returns:
            function class

        """
        functions = [func for func in self._functions if func.stream == stream and func.function == function]

        if len(functions) == 0:
            return None
        if len(functions) > 1:
            raise ValueError(f"More than one function found for S{stream:02}F{function:02}: {functions}")

        return functions[0]

    def decode(self, message: secsgem.common.Message) -> SecsStreamFunction:
        """Get object of decoded stream and function class, or None if no class is available.

        Args:
            message: message to get object for

        Returns:
            matching stream and function object

        """
        if message is None:
            raise ValueError("Decoding failed, missing message")

        func = self.function(message.header.stream, message.header.function)
        if func is None:
            raise ValueError("Decoding failed, invalid message")

        if isinstance(message.data, SecsStreamFunction):
            return message.data

        function = func()
        function.decode(message.data)

        return function



"""SECS dynamic variable type."""

from .array import Array
from .base import Base
from .binary import Binary
from .boolean import Boolean
from .f4 import F4
from .f8 import F8
from .i1 import I1
from .i2 import I2
from .i4 import I4
from .i8 import I8
from .string import String
from .u1 import U1
from .u2 import U2
from .u4 import U4
from .u8 import U8


class Dynamic(Base):
    """Variable with interchangable type."""

    def __init__(self, types, value=None, count=-1):
        """Initialize a dynamic secs variable.

        :param types: list of supported types, default first. empty means all types are support, String default
        :type types: list of :class:`secsgem.secs.variables.Base` classes
        :param value: initial value
        :type value: various
        :param count: max number of items in type
        :type count: integer
        """
        super().__init__()

        self.value = None

        self.types = types
        self.count = count
        if value is not None:
            self.set(value)

    def __repr__(self):
        """Generate textual representation for an object of this class."""
        return self.value.__repr__()

    def __len__(self):
        """Get the length."""
        return self.value.__len__()

    def __getitem__(self, key):
        """Get an item using the indexer operator."""
        return self.value.__getitem__(key)

    def __setitem__(self, key, item):
        """Set an item using the indexer operator."""
        self.value.__setitem__(key, item)

    def __eq__(self, other):
        """Check equality with other object."""
        if isinstance(other, Dynamic):
            return other.value.value == self.value.value
        if isinstance(other, Base):
            return other.value == self.value.value
        if isinstance(other, list):
            return other == self.value.value
        if isinstance(other, (bytes, str)) and isinstance(self.value.value, (bytes, str)):
            return str(other) == str(self.value.value)

        return [other] == self.value.value

    def __hash__(self):
        """Get data item for hashing."""
        if isinstance(self.value.value, list):
            return hash(self.value.value[0])
        return hash(self.value.value)

    def __type_supported(self, typ):
        if not self.types:
            return True

        if typ in self.types:
            return True

        return False

    @property
    def preferred_type(self):
        """Get the preferred type."""
        types = []
        for typ in self.types:
            types += typ.preferred_types

        return types[0]

    def set(self, value):
        """Set the internal value to the provided value.

        In doubt provide the variable wrapped in the matching :class:`secsgem.secs.variables.Base` class,
        to avoid confusion.

        Example:
            >>> import secsgem.secs
            >>>
            >>> var = secsgem.secs.variables.Dynamic([secsgem.secs.variables.String,
            ...                                       secsgem.secs.variables.U1])
            >>> var.set(secsgem.secs.variables.U1(10))
            >>> var
            <U1 10 >

        If no type is provided the default type is used which might not be the expected type.

        :param value: new value
        :type value: various
        """
        if isinstance(value, Base):
            if isinstance(value, Dynamic):
                if not isinstance(value.value, tuple(self.types)) and self.types:
                    raise ValueError(
                        f"Unsupported type {value.value.__class__.__name__} "
                        f"for this instance of Dynamic, allowed {self.types}")

                self.value = value.value
            else:
                if not isinstance(value, tuple(self.types)) and self.types:
                    raise ValueError(
                        f"Unsupported type {value.__class__.__name__} "
                        f"for this instance of Dynamic, allowed {self.types}")

                self.value = value
        else:
            matched_type = self._match_type(value)

            if matched_type is None:
                raise ValueError(
                    f'Value "{value}" of type {value.__class__.__name__} not valid for SecsDynamic with {self.types}')

            self.value = matched_type(count=self.count)
            self.value.set(value)

    def get(self):
        """Return the internal value.

        :returns: internal value
        :rtype: various
        """
        if self.value is not None:
            return self.value.get()

        return None

    def encode(self):
        """Encode the value to secs data.

        :returns: encoded data bytes
        :rtype: string
        """
        return self.value.encode()

    def decode(self, data, start=0):
        """Decode the secs byte data to the value.

        :param data: encoded data bytes
        :type data: string
        :param start: start position of value the data
        :type start: integer
        :returns: new start position
        :rtype: integer
        """
        (_, format_code, _) = self.decode_item_header(data, start)

        format_codes = {
            Array.format_code: Array,
            Binary.format_code: Binary,
            Boolean.format_code: Boolean,
            String.format_code: String,
            I8.format_code: I8,
            I1.format_code: I1,
            I2.format_code: I2,
            I4.format_code: I4,
            F8.format_code: F8,
            F4.format_code: F4,
            U8.format_code: U8,
            U1.format_code: U1,
            U2.format_code: U2,
            U4.format_code: U4,
        }

        if format_code not in format_codes or not self.__type_supported(format_codes[format_code]):
            raise ValueError(
                f"Unsupported format {format_code} for this instance of Dynamic, allowed {self.types}")

        typ = format_codes[format_code]
        if typ == Array:
            self.value = Array(ANYVALUE)
        else:
            self.value = typ(count=self.count)

        return self.value.decode(data, start)

    def _match_type(self, value):
        var_types = self.types
        # if no types are set use internal order
        if not self.types:
            var_types = [Boolean, U1, U2, U4, U8, I1, I2, I4,
                         I8, F4, F8, String, Binary]

        # first try to find the preferred type for the kind of value
        for var_type in var_types:
            if isinstance(value, tuple(var_type.preferred_types)) and var_type(count=self.count).supports_value(value):
                return var_type

        # when no preferred type was found, then try to match any available type
        for var_type in var_types:
            if var_type(count=self.count).supports_value(value):
                return var_type

        return None

    @property
    def is_dynamic(self) -> bool:
        """Check if this instance is Dynamic or derived."""
        return True


class ANYVALUE(Dynamic):
    """Dummy data item for generation of unknown types.

    :Types:
       - :class:`Array <secsgem.secs.variables.Array>`
       - :class:`Binary <secsgem.secs.variables.Binary>`
       - :class:`Boolean <secsgem.secs.variables.Boolean>`
       - :class:`String <secsgem.secs.variables.String>`
       - :class:`I8 <secsgem.secs.variables.I8>`
       - :class:`I1 <secsgem.secs.variables.I1>`
       - :class:`I2 <secsgem.secs.variables.I2>`
       - :class:`I4 <secsgem.secs.variables.I4>`
       - :class:`F8 <secsgem.secs.variables.F8>`
       - :class:`F4 <secsgem.secs.variables.F4>`
       - :class:`U8 <secsgem.secs.variables.U8>`
       - :class:`U1 <secsgem.secs.variables.U1>`
       - :class:`U2 <secsgem.secs.variables.U2>`
       - :class:`U4 <secsgem.secs.variables.U4>`

    """

    def __init__(self, value=None):
        """Initialize an ANYVALUE variable.

        :param value: value of the variable
        """
        self.name = self.__class__.__name__

        super().__init__([Array, Boolean, U1, U2, U4, U8, I1, I2, I4, I8, F4, F8, String, Binary], value=value)



"""Class for stream 01 function 16."""

from secsgem.secs.data_items import OFLACK
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS01F16(SecsStreamFunction):
    """offline acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F16
        OFLACK: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F16(secsgem.secs.data_items.OFLACK.ACK)
        S1F16
          <B 0x0> .

    Data Items:
        - :class:`OFLACK <secsgem.secs.data_items.OFLACK>`

    """

    _stream = 1
    _function = 16

    _data_format = OFLACK

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 05 function 17."""

from secsgem.secs.data_items import EXID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS05F17(SecsStreamFunction):
    """exception recover abort - request.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F17
        EXID: A[20]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F17("EX123")
        S5F17 W
          <A "EX123"> .

    Data Items:
        - :class:`EXID <secsgem.secs.data_items.EXID>`

    """

    _stream = 5
    _function = 17

    _data_format = EXID

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""State Models capability."""
from __future__ import annotations

import typing

import secsgem.common

from .capability import Capability
from .collection_event import CollectionEventId
from .communication_state_machine import CommunicationState
from .control_state_machine import ControlState, ControlStateMachine
from .handler import GemHandler

if typing.TYPE_CHECKING:
    import secsgem.secs


class StateModelsCapability(GemHandler, Capability):
    """State Models capability."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize capability."""
        super().__init__(*args, **kwargs)

        self._control_state = ControlStateMachine(self._initial_control_state, self._initial_online_control_state)

        # 3, send S01E01
        self._control_state.attempt_online.events.enter.register(self._on_control_state_attempt_online)

        # 7, send collection event
        self._control_state.transition("initial_online_local").events.called.register(
            self._on_control_state_initial_online_local,
        )

        # 8, send collection event
        self._control_state.transition("switch_online_local").events.called.register(
            self._on_control_state_initial_online_local,
        )

        # 8, send collection event
        self._control_state.transition("initial_online_remote").events.called.register(
            self._on_control_state_initial_online_remote,
        )

        # 9, send collection event
        self._control_state.transition("switch_online_remote").events.called.register(
            self._on_control_state_initial_online_remote,
        )

        self._control_state.start()

    @property
    def control_state(self) -> ControlStateMachine:
        """Get control state."""
        return self._control_state

    def _on_control_state_attempt_online(self, _):
        if self._communication_state.current != CommunicationState.COMMUNICATING:
            self._control_state.attempt_online_fail_host_offline()
            return

        response = self.are_you_there()

        if response is None:
            self._control_state.attempt_online_fail_host_offline()
            return

        if response.header.stream != 1 or response.header.function != 2:
            self._control_state.attempt_online_fail_host_offline()
            return

        self._control_state.attempt_online_success()

    def _on_control_state_initial_online_local(self, _):
        self.trigger_collection_events([CollectionEventId.CONTROL_STATE_LOCAL.value])

    def _on_control_state_initial_online_remote(self, _):
        self.trigger_collection_events([CollectionEventId.CONTROL_STATE_REMOTE.value])

    def control_switch_online(self):
        """Operator switches to online control state."""
        self._control_state.switch_online()

    def control_switch_offline(self):
        """Operator switches to offline control state."""
        self._control_state.switch_offline()
        self.trigger_collection_events([CollectionEventId.EQUIPMENT_OFFLINE.value])

    def control_switch_online_local(self):
        """Operator switches to the local online control state."""
        self._control_state.switch_online_local()

    def control_switch_online_remote(self):
        """Operator switches to the local online control state."""
        self._control_state.switch_online_remote()

    def _on_s01f15(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 1, Function 15, Request offline.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler, message  # unused parameters

        oflack = 0

        if self._control_state.current in [ControlState.ONLINE, ControlState.ONLINE_LOCAL, ControlState.ONLINE_REMOTE]:
            self._control_state.remote_offline()
            self.trigger_collection_events([CollectionEventId.EQUIPMENT_OFFLINE.value])

        return self.stream_function(1, 16)(oflack)

    def _on_s01f17(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 1, Function 17, Request online.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler, message  # unused parameters

        onlack = 1

        if self._control_state.current == ControlState.HOST_OFFLINE:
            self._control_state.remote_online()
            onlack = 0
        elif self._control_state.current in [ControlState.ONLINE,
                                             ControlState.ONLINE_LOCAL,
                                             ControlState.ONLINE_REMOTE]:
            onlack = 2

        return self.stream_function(1, 18)(onlack)

    def _get_control_state_id(self) -> int:
        """Get id of the control state for the current control state.

        :returns: control state
        :rtype: integer
        """
        if self._control_state.current == ControlState.EQUIPMENT_OFFLINE:
            return 1
        if self._control_state.current == ControlState.ATTEMPT_ONLINE:
            return 2
        if self._control_state.current == ControlState.HOST_OFFLINE:
            return 3
        if self._control_state.current == ControlState.ONLINE_LOCAL:
            return 4
        if self._control_state.current == ControlState.ONLINE_REMOTE:
            return 5

        return -1



"""DUTMS data item."""
from .. import variables
from .base import DataItemBase


class DUTMS(DataItemBase):
    """Die units of measure.

    :Type: :class:`String <secsgem.secs.variables.String>`

    **Used In Function**
        - :class:`SecsS12F01 <secsgem.secs.functions.SecsS12F01>`
        - :class:`SecsS12F04 <secsgem.secs.functions.SecsS12F04>`

    """

    __type__ = variables.String



"""Class for stream 02 function 18."""

from secsgem.secs.data_items import TIME
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F18(SecsStreamFunction):
    """date and time - data.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F18
        TIME: A[32]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F18("160816205942")
        S2F18
          <A "160816205942"> .

    Data Items:
        - :class:`TIME <secsgem.secs.data_items.TIME>`

    """

    _stream = 2
    _function = 18

    _data_format = TIME

    _to_host = True
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



# pylint: disable=non-parent-init-called
"""Data item base class."""
from __future__ import annotations

from .. import variables


class DataItemMeta(type):
    """Meta class for data items."""

    def __new__(cls, name, bases, attrs):
        """Meta class creation."""
        if name != "DataItemBase":
            bases += (attrs["__type__"], )
        return type.__new__(cls, name, bases, attrs)


class DataItemBase(metaclass=DataItemMeta):
    """Base class for data items.

    It provides type and output handling.
    """

    __type__: type[variables.Base] | None = None
    __allowedtypes__: list[type[variables.Base]] | None = None
    __count__ = -1

    def __init__(self, value=None):
        """Initialize a data item.

        :param value: Value of the data item
        """
        self.name = self.__class__.__name__

        if self.__type__ is variables.Dynamic:
            self.__type__.__init__(self, self.__allowedtypes__, value, self.__count__)
        else:
            self.__type__.__init__(self, value, self.__count__)

    @property
    def typ(self) -> type[variables.Base] | None:
        """Get the configured type."""
        return self.__type__

    @classmethod
    def get_format(cls, showname=True) -> str:
        """Format the contents as a string.

        Args:
            showname: Display the real class name when True

        Returns:
            Formatted value string

        """
        clsname = format(cls.__name__) if showname else "DATA"

        if cls.__type__ is variables.Dynamic:
            if cls.__count__ > 0:
                return f"{clsname}: {'/'.join([x.text_code for x in cls.__allowedtypes__])}[{cls.__count__}]"

            return f"{clsname}: {'/'.join([x.text_code for x in cls.__allowedtypes__])}"

        if cls.__count__ > 0:
            return f"{clsname}: {cls.__type__.text_code}[{cls.__count__}]"

        return f"{clsname}: {cls.__type__.text_code}"



"""Capability base class."""
from __future__ import annotations

import abc
import typing

if typing.TYPE_CHECKING:
    import secsgem.secs

    from .alarm import Alarm
    from .collection_event import CollectionEventId
    from .data_value import DataValue
    from .status_variable import StatusVariable


class Capability(abc.ABC):  # pylint: disable=too-few-public-methods
    """Capability base class for GEM equipment."""

    @property
    @abc.abstractmethod
    def _initial_control_state(self) -> str:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _initial_online_control_state(self) -> str:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _alarms(self) -> dict[int | str, Alarm]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _status_variables(self) -> dict[int | str, StatusVariable]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _data_values(self) -> dict[int | str, DataValue]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _time_format(self) -> int:
        raise NotImplementedError

    @_time_format.setter
    @abc.abstractmethod
    def _time_format(self, value: int):
        raise NotImplementedError

    @abc.abstractmethod
    def _get_sv_value(self, status_variable: StatusVariable) -> secsgem.secs.variables.Base:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_dv_value(self, data_value: DataValue) -> secsgem.secs.variables.Base:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_control_state_id(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_events_enabled(self) -> list[int | str]:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_alarms_enabled(self) -> list[int | str]:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_alarms_set(self) -> list[int | str]:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_clock(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def trigger_collection_events(self, ceids: list[int | str | CollectionEventId]):
        """Triggers the supplied collection events.

        Args:
            ceids: List of collection events

        """
        raise NotImplementedError



"""MHEAD data item."""
from .. import variables
from .base import DataItemBase


class MHEAD(DataItemBase):
    """SECS message header.

    :Type: :class:`Binary <secsgem.secs.variables.Binary>`
    :Length: 10

    **Used In Function**
        - :class:`SecsS09F01 <secsgem.secs.functions.SecsS09F01>`
        - :class:`SecsS09F03 <secsgem.secs.functions.SecsS09F03>`
        - :class:`SecsS09F05 <secsgem.secs.functions.SecsS09F05>`
        - :class:`SecsS09F07 <secsgem.secs.functions.SecsS09F07>`
        - :class:`SecsS09F11 <secsgem.secs.functions.SecsS09F11>`

    """

    __type__ = variables.Binary
    __count__ = 10



"""Handler for GEM host."""
from __future__ import annotations

import collections

import secsgem.common
import secsgem.secs

from .handler import GemHandler


class GemHostHandler(GemHandler):
    """Baseclass for creating host models. Inherit from this class and override required functions."""

    def __init__(self, settings: secsgem.common.Settings):
        """Initialize a gem host handler.

        Args:
            settings: communication settings

        """
        GemHandler.__init__(self, settings)

        self.is_host = True

        self.report_subscriptions: dict[int | str, list[int | str]] = {}

    def clear_collection_events(self) -> None:
        """Clear all collection events."""
        self._logger.info("Clearing collection events")

        # clear subscribed reports
        self.report_subscriptions = {}

        # disable all ceids
        self.disable_ceids()

        # delete all reports
        self.disable_ceid_reports()

    def subscribe_collection_event(self,
                                   ceid: int | str,
                                   dvs: list[int | str],
                                   report_id: int | str | None = None):
        """Subscribe to a collection event.

        :param ceid: ID of the collection event
        :type ceid: integer
        :param dvs: DV IDs to add for collection event
        :type dvs: list of integers
        :param report_id: optional - ID for report, autonumbering if None
        :type report_id: integer
        """
        self._logger.info("Subscribing to collection event %s", ceid)

        if report_id is None:
            report_id = self._report_id_counter
            self._report_id_counter += 1

        # note subscribed reports
        self.report_subscriptions[report_id] = dvs

        # create report
        self.send_and_waitfor_response(self.stream_function(2, 33)(
            {"DATAID": 0, "DATA": [{"RPTID": report_id, "VID": dvs}]}))

        # link event report to collection event
        self.send_and_waitfor_response(self.stream_function(2, 35)(
            {"DATAID": 0, "DATA": [{"CEID": ceid, "RPTID": [report_id]}]}))

        # enable collection event
        self.send_and_waitfor_response(self.stream_function(2, 37)({"CEED": True, "CEID": [ceid]}))

    def send_remote_command(self,
                            rcmd: int | str,
                            params: list[str]):
        """Send a remote command.

        :param rcmd: Name of command
        :type rcmd: string
        :param params: DV IDs to add for collection event
        :type params: list of strings
        """
        self._logger.info("Send RCMD %s", rcmd)

        s2f41 = self.stream_function(2, 41)()
        s2f41.RCMD = rcmd
        if isinstance(params, list):
            for param in params:
                s2f41.PARAMS.append({"CPNAME": param[0], "CPVAL": param[1]})
        elif isinstance(params, collections.OrderedDict):
            for param in params:
                s2f41.PARAMS.append({"CPNAME": param, "CPVAL": params[param]})

        # send remote command
        return self.settings.streams_functions.decode(self.send_and_waitfor_response(s2f41))

    def delete_process_programs(self,
                                ppids: list[int | str]):
        """Delete a list of process program.

        :param ppids: Process programs to delete
        :type ppids: list of strings
        """
        self._logger.info("Delete process programs %s", ppids)

        # send remote command
        return self.settings.streams_functions.decode(
            self.send_and_waitfor_response(self.stream_function(7, 17)(ppids))).get()

    def get_process_program_list(self) -> secsgem.secs.SecsStreamFunction:
        """Get process program list."""
        self._logger.info("Get process program list")

        # send remote command
        return self.settings.streams_functions.decode(
            self.send_and_waitfor_response(self.stream_function(7, 19)())).get()

    def go_online(self) -> str | None:
        """Set control state to online."""
        self._logger.info("Go online")

        # send remote command
        resp = self.settings.streams_functions.decode(self.send_and_waitfor_response(self.stream_function(1, 17)()))
        if resp is None:
            return None

        return resp.get()

    def go_offline(self) -> str | None:
        """Set control state to offline."""
        self._logger.info("Go offline")

        # send remote command
        return self.settings.streams_functions.decode(
            self.send_and_waitfor_response(self.stream_function(1, 15)())).get()

    def enable_alarm(self, alid: int | str):
        """Enable alarm.

        :param alid: alarm id to enable
        :type alid: :class:`secsgem.secs.dataitems.ALID`
        """
        self._logger.info("Enable alarm %d", alid)

        return self.settings.streams_functions.decode(self.send_and_waitfor_response(self.stream_function(5, 3)(
            {"ALED": secsgem.secs.data_items.ALED.ENABLE, "ALID": alid}))).get()

    def disable_alarm(self, alid: int | str):
        """Disable alarm.

        :param alid: alarm id to disable
        :type alid: :class:`secsgem.secs.dataitems.ALID`
        """
        self._logger.info("Disable alarm %d", alid)

        return self.settings.streams_functions.decode(self.send_and_waitfor_response(self.stream_function(5, 3)(
            {"ALED": secsgem.secs.data_items.ALED.DISABLE, "ALID": alid}))).get()

    def list_alarms(self,
                    alids: list[int | str] | None = None):
        """List alarms.

        :param alids: alarms to list details for
        :type alids: array of int/str
        """
        if alids is None:
            alids = []
            self._logger.info("List all alarms")
        else:
            self._logger.info("List alarms %s", alids)

        return self.settings.streams_functions.decode(
            self.send_and_waitfor_response(self.stream_function(5, 5)(alids))).get()

    def list_enabled_alarms(self):
        """List enabled alarms."""
        self._logger.info("List all enabled alarms")

        return self.settings.streams_functions.decode(
            self.send_and_waitfor_response(self.stream_function(5, 7)())).get()

    def _on_alarm_received(self, handler, alarm_id, alarm_code, alarm_text):
        del handler, alarm_id, alarm_code, alarm_text  # unused variables
        return secsgem.secs.data_items.ACKC5.ACCEPTED

    def _on_s05f01(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 5, Function 1, Alarm request.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        s5f1 = self.settings.streams_functions.decode(message)

        result = self._callback_handler.alarm_received(handler, s5f1.ALID, s5f1.ALCD, s5f1.ALTX)

        self.events.fire("alarm_received", {"code": s5f1.ALCD, "alid": s5f1.ALID, "text": s5f1.ALTX,
                                            "handler": self.protocol, "peer": self})

        return self.stream_function(5, 2)(result)

    def _on_s06f11(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 6, Function 11, Establish Communication Request.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler  # unused parameters

        function = self.settings.streams_functions.decode(message)

        for report in function.RPT:
            report_dvs = self.report_subscriptions[report.RPTID.get()]
            report_values = report.V.get()

            values = []

            for index, data_value_id in enumerate(report_dvs):
                values.append({"dvid": data_value_id,
                               "value": report_values[index]})

            data = {"ceid": function.CEID, "rptid": report.RPTID, "values": values,
                    "handler": self.protocol, "peer": self}
            self.events.fire("collection_event_received", data)

        return self.stream_function(6, 12)(0)

    def _on_terminal_received(self, handler, terminal_id, text):
        del handler, terminal_id, text  # unused variables
        return secsgem.secs.data_items.ACKC10.ACCEPTED

    def _on_s10f01(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 10, Function 1, Terminal Request.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        s10f1 = self.settings.streams_functions.decode(message)

        result = self._callback_handler.terminal_received(handler, s10f1.TID, s10f1.TEXT)
        self.events.fire("terminal_received", {"text": s10f1.TEXT, "terminal": s10f1.TID, "handler": self.protocol,
                                               "peer": self})

        return self.stream_function(10, 2)(result)



"""Class for stream 02 function 33."""

from secsgem.secs.data_items import DATAID, RPTID, VID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F33(SecsStreamFunction):
    """define report.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F33
        {
            DATAID: U1/U2/U4/U8/I1/I2/I4/I8/A
            DATA: [
                {
                    RPTID: U1/U2/U4/U8/I1/I2/I4/I8/A
                    VID: [
                        DATA: U1/U2/U4/U8/I1/I2/I4/I8/A
                        ...
                    ]
                }
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F33({"DATAID": 1, "DATA": [{"RPTID": 1000, "VID": [12, 1337]},
        ...     {"RPTID": 1001, "VID": [1, 2355]}]})
        S2F33 W
          <L [2]
            <U1 1 >
            <L [2]
              <L [2]
                <U2 1000 >
                <L [2]
                  <U1 12 >
                  <U2 1337 >
                >
              >
              <L [2]
                <U2 1001 >
                <L [2]
                  <U1 1 >
                  <U2 2355 >
                >
              >
            >
          > .

    Data Items:
        - :class:`DATAID <secsgem.secs.data_items.DATAID>`
        - :class:`RPTID <secsgem.secs.data_items.RPTID>`
        - :class:`VID <secsgem.secs.data_items.VID>`

    """

    _stream = 2
    _function = 33

    _data_format = [
        DATAID,
        [
            [
                RPTID,
                [VID]
            ]
        ]
    ]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = True



"""Wrapper for GEM status variable."""
from __future__ import annotations

import enum

import secsgem.secs


class StatusVariableId(enum.Enum):
    """Default IDs for status variables."""
    CLOCK = 1001
    CONTROL_STATE = 1002
    EVENTS_ENABLED = 1003
    ALARMS_ENABLED = 1004
    ALARMS_SET = 1005


class StatusVariable:  # pylint: disable=too-few-public-methods
    """Status variable definition."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        svid: int | str | StatusVariableId,
        name: str,
        unit: str,
        value_type: secsgem.secs.variables.Base,
        use_callback: bool = True,
        **kwargs,
    ):
        """Initialize a status variable.

        You can manually set the secs-type of the id with the 'id_type' keyword argument.

        Custom parameters can be set with the keyword arguments,
        they will be passed to the GemEquipmentHandlers callback
        :func:`secsgem.gem.equipmenthandler.GemEquipmentHandler.on_sv_value_request`.

        If use_callbacks is disabled, you can set the value with the value property.

        Args:
            svid: ID of the status variable
            name: long name of the status variable
            unit: unit (see SEMI E5, Units of Measure)
            value_type: type of the status variable
            use_callback: use the GemEquipmentHandler callbacks to get variable (True) or use internal value
            kwargs: extra class attributes

        """
        self.svid = svid if not isinstance(svid, StatusVariableId) else svid.value
        self.name = name
        self.unit = unit
        self.value_type = value_type
        self.use_callback = use_callback
        self.value = 0

        self.id_type: type[secsgem.secs.variables.Base]

        if isinstance(self.svid, int):
            self.id_type = secsgem.secs.variables.U4
        else:
            self.id_type = secsgem.secs.variables.String

        for key, value in kwargs.items():
            setattr(self, key, value)



"""RCMD data item."""
from .. import variables
from .base import DataItemBase


class RCMD(DataItemBase):
    """Remote command.

    :Types:
       - :class:`U1 <secsgem.secs.variables.U1>`
       - :class:`I1 <secsgem.secs.variables.I1>`
       - :class:`String <secsgem.secs.variables.String>`

    **Used In Function**
        - :class:`SecsS02F21 <secsgem.secs.functions.SecsS02F21>`
        - :class:`SecsS02F41 <secsgem.secs.functions.SecsS02F41>`
        - :class:`SecsS02F49 <secsgem.secs.functions.SecsS02F49>`

    """

    __type__ = variables.Dynamic
    __allowedtypes__ = [
        variables.U1,
        variables.I1,
        variables.String
    ]



"""Class for stream 05 function 03."""

from secsgem.secs.data_items import ALED, ALID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS05F03(SecsStreamFunction):
    """en-/disable alarm - send.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F03
        {
            ALED: B[1]
            ALID: U1/U2/U4/U8/I1/I2/I4/I8
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F03({"ALED": secsgem.secs.data_items.ALED.ENABLE, "ALID": 100})
        S5F3
          <L [2]
            <B 0x80>
            <U1 100 >
          > .

    Data Items:
        - :class:`ALED <secsgem.secs.data_items.ALED>`
        - :class:`ALID <secsgem.secs.data_items.ALID>`

    """

    _stream = 5
    _function = 3

    _data_format = [
        ALED,
        ALID
    ]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 12 function 18."""

from secsgem.secs.data_items import BINLT, IDTYP, MID, XYPOS
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS12F18(SecsStreamFunction):
    """map data type 3.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F18
        {
            MID: A/B[80]
            IDTYP: B[1]
            DATA: [
                {
                    XYPOS: I1/I2/I4/I8[2]
                    BINLT: U1/A
                }
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F18({
        ...     "MID": "materialID",
        ...     "IDTYP": secsgem.secs.data_items.IDTYP.WAFER,
        ...     "DATA": [
        ...         {"XYPOS": [1, 2], "BINLT": [1, 2, 3, 4]},
        ...         {"XYPOS": [3, 4], "BINLT": [5, 6, 7, 8]}]})
        S12F18
          <L [3]
            <A "materialID">
            <B 0x0>
            <L [2]
              <L [2]
                <I1 1 2 >
                <U1 1 2 3 4 >
              >
              <L [2]
                <I1 3 4 >
                <U1 5 6 7 8 >
              >
            >
          > .

    Data Items:
        - :class:`MID <secsgem.secs.data_items.MID>`
        - :class:`IDTYP <secsgem.secs.data_items.IDTYP>`
        - :class:`XYPOS <secsgem.secs.data_items.XYPOS>`
        - :class:`BINLT <secsgem.secs.data_items.BINLT>`

    """

    _stream = 12
    _function = 18

    _data_format = [
        MID,
        IDTYP,
        [
            [
                XYPOS,
                BINLT
            ]
        ]
    ]

    _to_host = False
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = True



"""Class for stream 09 function 13."""

from secsgem.secs.data_items import EDID, MEXP
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS09F13(SecsStreamFunction):
    """conversation timeout.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS09F13
        {
            MEXP: A[6]
            EDID: U1/U2/U4/U8/I1/I2/I4/I8/A/B
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS09F13({"MEXP": "S01E01", "EDID": "data"})
        S9F13
          <L [2]
            <A "S01E01">
            <A "data">
          > .

    Data Items:
        - :class:`MEXP <secsgem.secs.data_items.MEXP>`
        - :class:`EDID <secsgem.secs.data_items.EDID>`

    """

    _stream = 9
    _function = 13

    _data_format = [
        MEXP,
        EDID
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Clock capability."""
from __future__ import annotations

import datetime

from dateutil.tz import tzlocal

from .capability import Capability
from .handler import GemHandler


class ClockCapability(GemHandler, Capability):
    """Clock capability on GEM equipment."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize capability."""
        super().__init__(*args, **kwargs)

        self.__time_format = 1

    @property
    def _time_format(self) -> int:
        return self.__time_format

    @_time_format.setter
    def _time_format(self, value: int):
        self.__time_format = value

    def _get_clock(self) -> str:
        """Get the clock depending on configured time format.

        Returns:
            time code

        """
        now = datetime.datetime.now(tzlocal())
        if self._time_format == 0:
            return now.strftime("%y%m%d%H%M%S")

        if self._time_format == 2:
            return now.isoformat()

        return now.strftime("%Y%m%d%H%M%S") + now.strftime("%f")[0:2]



"""Header for the hsms reject request."""

from .header import HsmsHeader, HsmsSType


class HsmsRejectReqHeader(HsmsHeader):
    """Header for Reject Request.

    Header for message with SType 7.
    """

    def __init__(self, system: int, s_type: HsmsSType, reason: int):
        """Initialize a hsms reject request.

        :param system: message ID
        :type system: integer
        :param s_type: s_type of rejected message
        :type s_type: integer
        :param reason: reason for rejection
        :type reason: integer

        Example:
            >>> import secsgem.hsms
            >>>
            >>> secsgem.hsms.HsmsRejectReqHeader(17, secsgem.hsms.HsmsSType.DESELECT_REQ, 4)
            HsmsRejectReqHeader({session_id:0xffff, stream:03, function:04, p_type:0x00, s_type:0x07, \
system:0x00000011, require_response:False})
        """
        super().__init__(system, 0xFFFF, s_type.value, reason, False, 0x00, HsmsSType.REJECT_REQ)



"""Wrapper for GEM collection event report."""
from __future__ import annotations

import secsgem.secs


class CollectionEventReport:  # pylint: disable=too-few-public-methods
    """Report definition for registered collection events."""

    def __init__(self,
                 rptid: int | str,
                 variables: list[int | str],
                 **kwargs):
        """Initialize a collection event report.

        You can manually set the secs-type of the id with the 'id_type' keyword argument.

        Args:
            rptid: ID of the report
            variables: long name of the collection event
            **kwargs: additional attributes for object

        """
        self.rptid = rptid
        self.vars = variables

        self.id_type: type[secsgem.secs.variables.Base]

        if isinstance(self.rptid, int):
            self.id_type = secsgem.secs.variables.U4
        else:
            self.id_type = secsgem.secs.variables.String

        for key, value in kwargs.items():
            setattr(self, key, value)



"""BINLT data item."""
from .. import variables
from .base import DataItemBase


class BINLT(DataItemBase):
    """Bin list.

    :Types:
       - :class:`U1 <secsgem.secs.variables.U1>`
       - :class:`String <secsgem.secs.variables.String>`

    **Used In Function**
        - :class:`SecsS12F07 <secsgem.secs.functions.SecsS12F07>`
        - :class:`SecsS12F09 <secsgem.secs.functions.SecsS12F09>`
        - :class:`SecsS12F11 <secsgem.secs.functions.SecsS12F11>`
        - :class:`SecsS12F14 <secsgem.secs.functions.SecsS12F14>`
        - :class:`SecsS12F16 <secsgem.secs.functions.SecsS12F16>`
        - :class:`SecsS12F18 <secsgem.secs.functions.SecsS12F18>`

    """

    __type__ = variables.Dynamic
    __allowedtypes__ = [
        variables.U1,
        variables.String
    ]



"""Class for stream 05 function 05."""

from secsgem.secs.data_items import ALID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS05F05(SecsStreamFunction):
    """list alarms - request.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F05
        [
            ALID: U1/U2/U4/U8/I1/I2/I4/I8
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F05([100, 200])
        S5F5 W
          <L [2]
            <U1 100 >
            <U1 200 >
          > .

    Data Items:
        - :class:`ALID <secsgem.secs.data_items.ALID>`

    """

    _stream = 5
    _function = 5

    _data_format = [ALID]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Thread that can be triggered and call a callback."""
from __future__ import annotations

import logging
import queue
import threading
import typing

if typing.TYPE_CHECKING:
    from .message import Block
    from .settings import Settings


class ProtocolDispatcher:  # pylint: disable=too-many-instance-attributes
    """Thread that calls a target function when a trigger was raised."""

    def __init__(
            self,
            receiver_target: typing.Callable,
            dispatcher_target: typing.Callable,
            settings: Settings,
    ) -> None:
        """Initialize thread object.

        Args:
            receiver_target: function to call when receiver triggered
            dispatcher_target: function to call when message available for dispatch
            settings: communication/protocol settings

        """
        self._receiver_target = receiver_target
        self._dispatcher_target = dispatcher_target
        self._settings = settings

        self._receiver_thread: threading.Thread | None = None
        self._dispatcher_thread: threading.Thread | None = None

        self._receiver_thread_trigger = threading.Event()
        self._dispatcher_thread_trigger = threading.Event()

        self._dispatch_queue: queue.Queue[tuple[object, Block]] = queue.Queue()

        self._stop_receiver_thread = False
        self._stop_dispatcher_thread = False

    def start(self):
        """Start the thread."""
        self._stop_receiver_thread = False
        self._stop_dispatcher_thread = False

        self._receiver_thread = threading.Thread(
            target=self._receiver_thread_function,
            args=(),
            name=self._settings.generate_thread_name("protocol_receiver"),
            daemon=True,
        )

        self._dispatcher_thread = threading.Thread(
            target=self._dispatcher_thread_function,
            args=(),
            name=self._settings.generate_thread_name("protocol_dispatcher"),
            daemon=True,
        )

        self._receiver_thread.start()
        self._dispatcher_thread.start()

    def stop(self):
        """Stop the thread."""
        if not self._receiver_thread.is_alive():
            return

        self._stop_receiver_thread = True
        self._receiver_thread_trigger.set()

        self._receiver_thread.join()

    def trigger_receiver(self):
        """Trigger the thread to call target function."""
        self._receiver_thread_trigger.set()

    def queue_block(self, source: object, block: Block):
        """Add a block to the dispatch queue and trigger dispatch thread.

        Args:
            source: source of the block
            block: new block

        """
        self._dispatch_queue.put((source, block))
        self._dispatcher_thread_trigger.set()

    def _receiver_thread_function(self):
        while not self._stop_receiver_thread:
            self._receiver_thread_trigger.wait()
            self._receiver_thread_trigger.clear()

            if self._stop_receiver_thread:
                continue

            try:
                self._receiver_target()
            except Exception as exc:  # pylint: disable=broad-exception-caught
                logging.warning("Exception in receiver callback, ignoring", exc_info=exc)

        self._stop_receiver_thread = False

    def _dispatcher_thread_function(self):
        while not self._stop_dispatcher_thread:
            self._dispatcher_thread_trigger.wait()
            self._dispatcher_thread_trigger.clear()

            if self._stop_dispatcher_thread:
                continue

            while self._dispatch_queue.qsize() > 0:
                data = self._dispatch_queue.get()

                try:
                    self._dispatcher_target(*data)
                except Exception as exc:  # pylint: disable=broad-exception-caught
                    logging.warning("Exception in dispatcher callback, ignoring", exc_info=exc)

        self._stop_dispatcher_thread = False



"""Wrapper for GEM collection event link."""
from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from .collection_event import CollectionEvent
    from .collection_event_report import CollectionEventReport


class CollectionEventLink:
    """Representation for registered/linked collection event."""

    def __init__(self,
                 collection_event: CollectionEvent,
                 reports: list[CollectionEventReport],
                 **kwargs):
        """Initialize a collection event link.

        Args:
            collection_event: collection event
            reports: list of the linked reports
            **kwargs: additional attributes for object

        """
        self._collection_event = collection_event
        self._reports = reports
        self.enabled = False

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def collection_event(self) -> CollectionEvent:
        """Get the associated collection event."""
        return self._collection_event

    @property
    def reports(self):
        """Get list of the data values.

        Returns:
            List of linked reports

        """
        return self._reports



"""JIS X 0201 Codec required for JIS8 encoding of JIS8."""
import codecs

jis8_decoding_map = codecs.make_identity_dict(range(256))  # type: ignore[attr-defined]
jis8_decoding_map.update({
    0x005C: 0x00A5,  # Yen Sign
    0x007E: 0x203E,  # Overline
})

for i in range(0x00A1, 0x00E0):
    jis8_decoding_map[i] = i + 0xFEC0

jis8_encoding_map = codecs.make_encoding_map(jis8_decoding_map)  # type: ignore[attr-defined]


def _jis_x_0201_encode(data, errors="strict"):
    return codecs.charmap_encode(data, errors, jis8_encoding_map)


def _jis_x_0201_decode(data, errors="strict"):
    return codecs.charmap_decode(data, errors, jis8_decoding_map)


def _jis_x_0201_search(name):
    if name == "jis_8":
        return codecs.CodecInfo(encode=_jis_x_0201_encode, decode=_jis_x_0201_decode, name="jis_8")

    return None


# register the codec
codecs.register(_jis_x_0201_search)



"""Class for stream 02 function 50."""

from secsgem.secs.data_items import CPACK, CPNAME, HCACK
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F50(SecsStreamFunction):
    """Enhanced remote command - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F50
        {
            HCACK: B[1]
            PARAMS: [
                {
                    CPNAME: U1/U2/U4/U8/I1/I2/I4/I8/A
                    CPACK: B[1]
                }
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F50({
        ...     "HCACK": secsgem.secs.data_items.HCACK.INVALID_COMMAND,
        ...     "PARAMS": [
        ...         {"CPNAME": "PARAM1", "CPACK": secsgem.secs.data_items.CPACK.CPVAL_ILLEGAL_VALUE},
        ...         {"CPNAME": "PARAM2", "CPACK": secsgem.secs.data_items.CPACK.CPVAL_ILLEGAL_FORMAT}]})
        S2F50
          <L [2]
            <B 0x1>
            <L [2]
              <L [2]
                <A "PARAM1">
                <B 0x2>
              >
              <L [2]
                <A "PARAM2">
                <B 0x3>
              >
            >
          > .

    Data Items:
        - :class:`HCACK <secsgem.secs.data_items.HCACK>`
        - :class:`CPNAME <secsgem.secs.data_items.CPNAME>`
        - :class:`CPACK <secsgem.secs.data_items.CPACK>`

    """

    _stream = 2
    _function = 50

    _data_format = [
        HCACK,
        [
            [
                "PARAMS",
                CPNAME,
                CPACK
            ]
        ]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""EXRECVRA data item."""
from .. import variables
from .base import DataItemBase


class EXRECVRA(DataItemBase):
    """Exception recovery action.

    :Type: :class:`String <secsgem.secs.variables.String>`
    :Length: 40

    **Used In Function**
        - :class:`SecsS05F09 <secsgem.secs.functions.SecsS05F09>`
        - :class:`SecsS05F13 <secsgem.secs.functions.SecsS05F13>`

    """

    __type__ = variables.String
    __count__ = 40



"""Class for stream 07 function 05."""

from secsgem.secs.data_items import PPID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS07F05(SecsStreamFunction):
    """process program - request.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F05
        PPID: A/B[120]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F05("program")
        S7F5 W
          <A "program"> .

    Data Items:
        - :class:`PPID <secsgem.secs.data_items.PPID>`

    """

    _stream = 7
    _function = 5

    _data_format = PPID

    _to_host = True
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""HSMS settings class."""
from __future__ import annotations

import enum

import secsgem.common


class HsmsConnectMode(enum.Enum):
    """Hsms connect mode (active or passive)."""

    ACTIVE = 1
    PASSIVE = 2

    def __repr__(self) -> str:
        """String representation of object."""
        return "Active" if self == self.ACTIVE else "Passive"


class HsmsSettings(secsgem.common.Settings):
    """Settings for HSMS connection.

    These attributes can be initialized in the constructor and accessed as property.

    Example:
        >>> import secsgem.hsms
        >>>
        >>> settings = secsgem.hsms.HsmsSettings(device_type=secsgem.common.DeviceType.EQUIPMENT)
        >>> settings.device_type
        <DeviceType.EQUIPMENT: 0>
        >>> settings.address
        '127.0.0.1'

    """

    def __init__(self, **kwargs) -> None:
        """Initialize settings."""
        super().__init__(**kwargs)

        self._connect_mode = kwargs.get("connect_mode", HsmsConnectMode.ACTIVE)
        self._address = kwargs.get("address", "127.0.0.1")
        self._port = kwargs.get("port", 5000)

    @property
    def connect_mode(self) -> HsmsConnectMode:
        """Hsms connect mode.

        Default: HsmsConnectMode.ACTIVE
        """
        return self._connect_mode

    @property
    def address(self) -> str:
        """Remote (active) or local (passive) IP address.

        Default: "127.0.0.1"
        """
        return self._address

    @property
    def port(self) -> int:
        """TCP port of remote host.

        Default: 5000
        """
        return self._port

    def create_protocol(self) -> secsgem.common.Protocol:
        """Protocol class for this configuration."""
        from .protocol import HsmsProtocol  # pylint: disable=import-outside-toplevel

        return HsmsProtocol(self)

    def create_connection(self) -> secsgem.common.Connection:
        """Connection class for this configuration."""
        if self.connect_mode == HsmsConnectMode.ACTIVE:
            return secsgem.common.TcpClientConnection(self)
        return secsgem.common.TcpServerConnection(self)

    @property
    def name(self) -> str:
        """Name of this configuration."""
        return f"HSMS-{self.connect_mode}_{self.address}:{self.port}"

    @property
    def is_active(self) -> bool:
        """Check if connection is active."""
        return self.connect_mode == HsmsConnectMode.ACTIVE

    def generate_thread_name(self, functionality: str) -> str:
        """Generate a unique thread name for this configuration and a provided functionality.

        Args:
            functionality: name of the functionality to generate thread name for

        Returns:
            generated thread name

        """
        return f"secsgem_HSMS_{functionality}_{self.connect_mode}_{self.address}:{self.port}"


class ExistingProtocolSettings(HsmsSettings):
    """Settings for existing HSMS connection.

    These attributes can be initialized in the constructor and accessed as property.

    Example:
        >>> import secsgem.hsms
        >>>
        >>> settings = secsgem.hsms.HsmsSettings(device_type=secsgem.common.DeviceType.EQUIPMENT)
        >>> settings.device_type
        <DeviceType.EQUIPMENT: 0>

    """

    def __init__(self, **kwargs) -> None:
        """Initialize settings."""
        super().__init__(**kwargs)

        self._existing_protocol = kwargs.get("existing_protocol", None)

    @property
    def existing_protocol(self) -> secsgem.common.Protocol:
        """Existing protocol.

        Default: None
        """
        return self._existing_protocol

    def create_protocol(self) -> secsgem.common.Protocol:
        """Protocol class for this configuration."""
        return self.existing_protocol

    @property
    def name(self) -> str:
        """Name of this configuration."""
        return f"HSMS-{self.connect_mode}_{self.address}:{self.port}"



"""Class for stream 12 function 09."""

from secsgem.secs.data_items import BINLT, IDTYP, MID, STRP
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS12F09(SecsStreamFunction):
    """map data type 2 - send.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F09
        {
            MID: A/B[80]
            IDTYP: B[1]
            STRP: I1/I2/I4/I8[2]
            BINLT: U1/A
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F09({
        ...     "MID": "materialID",
        ...     "IDTYP": secsgem.secs.data_items.IDTYP.WAFER,
        ...     "STRP": [0, 1],
        ...     "BINLT": [1, 2, 3, 4, 5, 6]})
        S12F9 W
          <L [4]
            <A "materialID">
            <B 0x0>
            <I1 0 1 >
            <U1 1 2 3 4 5 6 >
          > .

    Data Items:
        - :class:`MID <secsgem.secs.data_items.MID>`
        - :class:`IDTYP <secsgem.secs.data_items.IDTYP>`
        - :class:`STRP <secsgem.secs.data_items.STRP>`
        - :class:`BINLT <secsgem.secs.data_items.BINLT>`

    """

    _stream = 12
    _function = 9

    _data_format = [
        MID,
        IDTYP,
        STRP,
        BINLT
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = True



"""SECS jis8 text variable type."""

import unicodedata

import secsgem.common.codec_jis_x_0201  # noqa: F401 pylint: disable=unused-import

from .base_text import BaseText


class JIS8(BaseText):
    """Secs type for string data.

    :param value: initial value
    :type value: string
    :param count: number of items this value
    :type count: integer
    """

    format_code = 0o21
    text_code = "J"
    preferred_types = [bytes, str]
    control_chars = "".join(chr(ch) for ch in range(256) if unicodedata.category(chr(ch))[0] == "C")
    coding = "jis_8"
